#ifndef s11n_SQLITE3_SERIALIZER_HPP_INCLUDED
#define s11n_SQLITE3_SERIALIZER_HPP_INCLUDED 1

#include <s11n.net/s11n/io/data_node_io.hpp> // data_node_serializer<> type
#include <s11n.net/s11n/io/sqlite3/sqlite3_util.hpp> // DBScope class


/**
   This string holds the text header prefixing all sqlite3
   databases.
*/
#define s11n_SQLITE_MAGIC_COOKIE "SQLite format 3"

/**
   s11n_SQLITE_WRITES_SQL defines whether the sqlite3_serializer
   knows how to output SQL to a stream.
*/
#define s11n_SQLITE_WRITES_SQL 1

/**
 s11n_SQLITE_WRITES_SQL_AND_DB is an internal expermiment/debugging tool. Don't
 rely on it.
*/
#define s11n_SQLITE_WRITES_SQL_AND_DB 0

namespace s11n {
namespace io {
namespace sqlite3 {


	/**
	   sqlite3_serializer is an sqlite3-based Serializer.

	   Due to limits of its database heritage, it cannot
	   deserialize from streams, and can only serialize
	   over streams as SQL.
	*/
	template <typename NodeType>
	class sqlite3_serializer : public data_node_serializer<NodeType>
	{
	public:
		typedef NodeType node_type;

		sqlite3_serializer()
		{
			this->magic_cookie( s11n_SQLITE_MAGIC_COOKIE );
		}

		virtual ~sqlite3_serializer() {}

		virtual bool serialize( const node_type & src, const std::string & dest )
		{
			if( dest.empty() ) return false;
#if s11n_SQLITE_WRITES_SQL_AND_DB
			return save_node_to_both<node_type>( src, dest );
#else
			DBScope dbw( dest ); // might throw
			save_node<node_type>( src, dbw.db ); // might throw
			return true;
#endif
		}

		virtual node_type * deserialize( const std::string & src )
		{
			// CERR << "sqlite3_serializer::deserialize("<<src<<")...\n";
			DBScope dbw( src ); // might throw
			return load_node<node_type>( dbw.db );
		}

		/**
		   Writes SQL code for src to dest and returns true.
		   Note that this function cannot know if the SQL code
		   is acceptable to the SQL engine, so it has no failure
		   case except that dest is not good.
		*/
		virtual bool serialize( const node_type & src, std::ostream & dest)
		{
#if s11n_SQLITE_WRITES_SQL
			if( ! dest.good() ) return false;
			save_node<node_type>( src, dest );
			return true;
#else
			throw s11n::s11n_exception( "sqlite3_serializer::serialize(node,ostream): this serializer cannot serialize to streams, only to files." );
			return false;
#endif // s11n_SQLITE_WRITES_SQL
		}

		/**
		   Always throws an exception, as this type cannot work
		   with streams.
		*/
		virtual node_type * deserialize( std::istream & src )
		{
			throw s11n::s11n_exception( "sqlite3_deserializer::deserialize(istream): this serializer cannot handle stream-based deserialization." );
			return 0;
		}

	};

} } } // namespace s11n::io::sqlite3



#endif // s11n_SQLITE3_SERIALIZER_HPP_INCLUDED
