
#include <s11n.net/s11n/io/sqlite3/sqlite3_util.hpp>
#include <fstream>
#include <vector>

namespace s11n { namespace io { namespace sqlite3 {


	DBScope::DBScope( std::string const & dbfile ) : db(0)
	{
        	int rc = sqlite3_open(dbfile.c_str(), &this->db);
	        if( rc )
        	{
			this->close();
	                throw std::runtime_error( std::string("Can't open database file "+dbfile+": ")+sqlite3_errmsg(this->db) );
        	}
	}

	int DBScope::exec( std::string const & sql, bool throwOnError )
	{
		int rc = this->db
			? sqlite3_exec( this->db, sql.c_str(), 0,0,0 )
			: -1;
		if( (0==rc) || ((0!=rc) && !throwOnError) ) return rc;
		// if( !rc || !throwOnError ) return rc;
		throw std::runtime_error( std::string("DBScope::exec(): SQL error: ")+sqlite3_errmsg(this->db)+"\nSQL="+sql  );
		return 0; // avoid warning from over-zealous compilers
	}

	void DBScope::close()
	{
		if( this->db )
		{
			sqlite3_close( this->db );
			this->db = 0;
		}
	}

	DBScope::~DBScope()
	{
		this->close();
	}


        size_t escape_string( std::string & ins )
        {
                std::string::size_type pos;
		const std::string to_esc("'");
                pos = ins.find_first_of( to_esc );
		if( std::string::npos == pos ) return 0;
		const std::string esc("'");
                size_t reps = 0;
                while( pos != std::string::npos )
                {
                        ins.insert( pos, esc );
                        ++reps;
                        pos = ins.find_first_of( to_esc, pos + esc.size() + 1 );
                }
                return reps;
        }


	int sql_to_db( std::istream & sql, ::sqlite3 * db, bool stopOnFirstError )
	{
		int rc = 0;
		std::string line;
		std::string buff;
		std::string errmsg;
		buff = "";
		while( std::getline( sql, line ) )
		{
			if( line.empty() ) continue;
			buff += line;
			if( sqlite3_complete( buff.c_str() ) )
			{
				rc = sqlite3_exec( db, buff.c_str(), 0, 0, 0 );
				buff = "";
				if( (0 != rc) && stopOnFirstError )
				{
					return rc;
				}
			}
			else
			{
				buff += '\n';
			}
		}
		if( ! buff.empty() )
		{
			rc = sqlite3_exec( db, buff.c_str(), 0, 0, 0 );
		}
		return rc;
}


	void get_cleanup_sql( std::ostream & sql)
	{
		sql << "DROP TABLE "<<s11n_SQLITE_TBL_T<<";\n";
		sql << "VACUUM;";

	}
	void get_drop_sql( std::ostream & sql )
	{
		sql //<< "BEGIN TRANSACTION dropping_stuff;\n"
			<< "DROP TABLE "<<s11n_SQLITE_TBL_N<<";\n"
			<< "DROP TABLE "<<s11n_SQLITE_TBL_P<<";\n"
			// << "COMMIT TRANSACTION dropping_stuff;\n"
			;
		sql << std::endl;
	}

	void get_create_sql( std::ostream & sql )
	{
 		//sql << "BEGIN TRANSACTION create_s11n_schema;\n";

		sql << "CREATE TABLE /* snodes */ "<<s11n_SQLITE_TBL_N<<"("
		    << "rowid INTEGER PRIMARY KEY AUTOINCREMENT,"
		    << "parent_id INTEGER, "
		    << "class, name);\n";

		sql << "CREATE TABLE /* properties */ "<<s11n_SQLITE_TBL_P<<"("
		    << "node_id INTEGER REFERENCES "<<s11n_SQLITE_TBL_N<<"(rowid)," /* REFERENCES does speed it up slighly. */
		    << "key TEXT,"
		    << "value TEXT);\n";

		sql << "CREATE TEMP TABLE "<<s11n_SQLITE_TBL_T<<"(key PRIMARY KEY ON CONFLICT REPLACE,value);\n"
		    << "INSERT INTO "<<s11n_SQLITE_TBL_T<<" values('current_node_id',0);\n"
			;

		// This trigger recursively deletes all sub-nodes and properties when a node is deleted.
		// We really shouldn't add it until the tables are populated, but that
		// would be too easy to forget to do.
		if( 0 ) {
			sql << "CREATE TRIGGER tr_cleanup_node BEFORE DELETE ON "<<s11n_SQLITE_TBL_N<<" BEGIN\n"
			    << "DELETE FROM "<<s11n_SQLITE_TBL_P<<" WHERE node_id = OLD.rowid;\n"
			    << "DELETE FROM "<<s11n_SQLITE_TBL_N<<" WHERE parent_id = OLD.rowid;\n"
			    << "END;\n";
		} else { // sqlite3 docs recommend this:
			sql << "CREATE TRIGGER tr_cleanup_node BEFORE DELETE ON "<<s11n_SQLITE_TBL_N<<"\n"
			    << "FOR EACH ROW BEGIN\n"
			    << "DELETE FROM "<<s11n_SQLITE_TBL_P<<" WHERE node_id = OLD.rowid;\n"
			    << "DELETE FROM "<<s11n_SQLITE_TBL_N<<" WHERE parent_id = OLD.rowid;\n"
			    << "END;\n";
		}

		sql << "CREATE INDEX ndx_nodes ON "<<s11n_SQLITE_TBL_N<<" (rowid,parent_id);\n";
		sql << "CREATE INDEX ndx_props ON "<<s11n_SQLITE_TBL_P<<" (node_id);\n";
		sql << "CREATE INDEX ndx_s11n_temp ON "<<s11n_SQLITE_TBL_T<<" (key) ON CONFLICT REPLACE;\n";

 		//sql << "COMMIT TRANSACTION create_s11n_schema;\n";

		sql << std::endl;
	}

	void get_header_sql( std::ostream & sql )
	{
		sql << "-- s11n::io::sqlite3::sqlite3_serializer\n";
		sql << "PRAGMA synchronous = OFF;\n"; /* Cuts write times by writes by 4-5x on my box. */

		sql << std::endl;
	}

#if 0 // wont't wrok because sqlite doesn't have the API to do this...
	static void sqfunc_read_file(
				     sqlite3_context *context,
				     int argc,
				     sqlite3_value **argv
				    )
	{
		int rc = 0;
		std::string fname;
		std::string line;
		std::string buff;
		std::string errmsg;
		for( int i = 0; (0 == rc) && (i < argc); ++i )
		{
			fname = static_cast<char const *>( sqlite3_value_text( argv[i] ) );
			std::ifstream in( fname.c_str() );
			if( ! in.good() )
			{
				errmsg = "Could not open file: "+fname;
				rc = 1;
				break;
			}
			buff = "";
			while( std::getline( in, line ) )
			{
				buff += "\n"+line;
				if( sqlite3_complete( buff.c_str() ) )
				{
					rc = sqlite3_exec( db, buff.c_str(), 0, 0, 0 );
					buff = "";
					if( 0 != rc ) break;
				}
			}
			if( ! buf.empty() )
			{
				rc = sqlite3_exec( db, buff.c_str(), 0, 0, 0 );
				buff = "";
			}
		}

		if( 0 != rc )
		{
			sqlite3_result_error( context, errmsg.empty()
					     ? sqlite3_errmsg(db)
					     : errmsg.c_str(),
					     SQLITE_TRANSIENT );
		}
	}
	int setup_functions( ::sqlite3 * db )
	{
		if( 0 ) return sqlite3_create_function(db, "read_file", -1, SQLITE_UTF8,
						       0, sqfunc_read_file, 0, 0);
		return 0;
	}
#endif // 1|0


	bool init_db( ::sqlite3 * db )
	{
		std::ostringstream sql;
		//setup_functions(db);
		get_header_sql( sql );
		get_drop_sql( sql );
		sqlite3_exec( db, sql.str().c_str() , NULL, 0, 0 );
		sql.str("");
		get_create_sql( sql );
		int rc = sqlite3_exec( db, sql.str().c_str() , NULL, 0, 0 );
		if( 1 && (0 != rc) )
		{
			CERR << "init_db() SQL error: "
			     << sqlite3_errmsg(db) << '\n';

		}
		return 0 == rc;
	}

} } } // namespaces


