#ifndef s11n_SQLITE3_UTIL_HPP_INCLUDED
#define s11n_SQLITE3_UTIL_HPP_INCLUDED 1

#include <iostream>
#include <fstream>
#include <sstream>
#include <stdexcept>

#include <sqlite3.h>

#include <s11n.net/s11n/s11n_config.hpp>
#include <s11n.net/s11n/exception.hpp>
#include <s11n.net/s11n/traits.hpp>

#if (s11n_S11N_LIBRARY_VERSION_HEX < 0x010201)
/* s11n < 1.2.1 */
#  error "The sqlite3 s11n add-on requires libs11n >= 1.2.1."
#endif

/**
   The s11n_SQLITE_TBL_xxx macros are names for the s11n db tables.
   Changing these breaks compatibility with older files, so don't
   do it often.
*/
#define s11n_SQLITE_TBL_T "t" /* tmp table name */
#define s11n_SQLITE_TBL_N "n" /* nodes table name */
#define s11n_SQLITE_TBL_P "p" /* properties table name */

#define SQLDEBUG if( 0 ) CERR << "DB: "

namespace s11n {
namespace io {
	/**
	   The sqlite3 namespace holds some code for using sqlite3
	   databases as s11n data stores. It is based upon the excellent
	   sqlite3 engine, available at http://sqlite.org.
	*/
namespace sqlite3 {


	/**
	   A utility type to simplify opening/closing of sqlite3
	   files.
	*/
	struct DBScope
	{
		/**
		   sqlite3_open()s infile. Throws a std::exception on error.
		*/
		DBScope( std::string const & infile );
		/**
		   Calls this->close().
		*/
		~DBScope();

		/**
		   DB handle. If you manually call sqlite3_close() on
		   this object then assign it to zero to ensure that
		   this->close() does not try to close it again.
		*/
		::sqlite3 * db;


		/**
		   Returns sqlite3_exec( this->db, sql.c_str(), 0,0,0 ),
		   or -1 if (! this->db).

		   If throwOnError is true then a non-0 return code will cause
		   an exception to be thrown.
		*/
		int exec( std::string const & sql, bool throwOnError = false );

		/**
		   Calls s1lite3_close() on this->db. If this->db is 0
		   then it does nothing. It should never be necessary
		   to manually call this unless you want to clean up
		   your db connection long before this object goes out
		   of scope.
		*/
		void close();
	private:
		DBScope( DBScope const & ); // Unimplemented. Noncopyable.
		DBScope & operator=( DBScope const & ); // Unimplemented. Noncopyable.
	};


	/**
	   If ins contains no apostrophes, this function does nothing
	   and returns zero, otherwise it tries to escape ins to make
	   it compliant for use in SQL code. That is, all apostrophes
	   are simply doubled. This algo is not particularly clever.
	   It is probably easily fooled.

	   Returns the number of translations made. If ins contains no
	   apostrophes then this function returns 0 and has no
	   side-effects.
	*/
	size_t escape_string( std::string & ins );


	/**
	   Sets up db for use as an s11n repository. Deletes any
	   existing s11n data in the db to enforce a one-db ==
	   one root snode policy.

	*/
	bool init_db( ::sqlite3 * db );


	/**
	   Reads SQL from the stream until EOF. To avoid buffering the
	   whole stream contents, it reads line-wise and uses
	   sqlite3_complete() to determine what constitutes a complete
	   statement. After reading a complete statement is executes
	   the code. Because it reads entire lines, several complete
	   statements on one line will be executed in one sqlite3_exec()
	   call.

	   If stopOnFirstError is true then it returns after the first
	   error result, otherwise it continues until the end and only
	   returns the result of the final SQL call.
	*/
	int sql_to_db( std::istream & sql, ::sqlite3 * db, bool stopOnFirstError );


	/**
	   The type sqlite3 uses for the automatic ROWID field.
	*/
	typedef long long int rowid_t;

	/**
	   Sends the "opening" SQL for creating the DB to the given
	   ostream.
	*/
	void get_header_sql( std::ostream & );

	void get_drop_sql( std::ostream & );
	void get_create_sql( std::ostream & );
	void get_cleanup_sql( std::ostream & );

	/**
	   Stores each property of src in the properties table of
	   db. node_id is assumed to be the sqlite ROWID of src.

	   Throws a std::exception on error.
	*/
	template <typename NodeType>
	void save_node_properties( NodeType const & src, ::sqlite3 * db, rowid_t node_id )
	{
		typedef s11n::node_traits<NodeType> NTR;
		typedef typename NTR::property_map_type::const_iterator PIT;
		PIT pit = NTR::properties(src).begin();
		PIT pet = NTR::properties(src).end();
		std::ostringstream sql;
		std::string pkey;
		std::string pval;
		size_t ndx = 0;
		int rc = 0;
		char * dberr = 0;
		for( ; pet != pit; ++pit )
		{
			pkey = (*pit).first;
			pval = (*pit).second;
			escape_string( pval );
			// todo: escape them!
			sql << "INSERT INTO "<<s11n_SQLITE_TBL_P<<" values("
			    <<node_id <<","
			    << "'" << pkey << "',"
			    << "'" << pval << "'"
			    << ");\n";
#define s11n_SQLITE_FLUSH_EVERY_N 20
#if s11n_SQLITE_FLUSH_EVERY_N // disable/enable periodic flushing...
			if( ++ndx == s11n_SQLITE_FLUSH_EVERY_N ) // rather arbitrary number.
			{ /** Every N properties we'll go ahead and
			      write, to make tracking down errors
			      simpler and avoid any potential
			      overflows in sqlite3's input limits.  It
			      also avoids us buffering too much data
			      if nodes contain lots of data in their
			      properties.
			  */
				ndx = 0;
				rc = sqlite3_exec( db, sql.str().c_str(), NULL, 0, &dberr );
				if( SQLITE_OK != rc )
				{
					throw s11n::s11n_exception( "Error execing SQL:\n"+sql.str()+"\nSQL error string="+dberr );
				}
				sql.str("");
			}
#else
			ndx = 0; // avoid compiler warning about unused var ndx.
#endif // s11n_SQLITE_FLUSH_EVERY_N
#undef s11n_SQLITE_FLUSH_EVERY_N
		}

		if( ! sql.str().empty() )
		{
			rc = sqlite3_exec( db, sql.str().c_str(), NULL, 0, &dberr );
			if( SQLITE_OK != rc )
			{
				throw s11n::s11n_exception( "Error execing SQL:\n"+sql.str()+"\nSQL error string="+dberr );
			}
		}
	}


	/**
	   See save_node_properties( NodeType, ::sqlite3, rowid_t ). This function
	   is similar, but generates SQL and sends it to dest.
	*/
	template <typename NodeType>
	void save_node_properties( NodeType const & src, std::ostream & dest )
	{
		typedef s11n::node_traits<NodeType> NTR;
		typedef typename NTR::property_map_type::const_iterator PIT;
		PIT pit = NTR::properties(src).begin();
		PIT pet = NTR::properties(src).end();
		std::string pkey;
		std::string pval;
		for( ; pet != pit; ++pit )
		{
			pkey = (*pit).first;
			pval = (*pit).second;
			escape_string( pval );
			dest << "INSERT INTO "<<s11n_SQLITE_TBL_P<<" values("
			     <<"(select value from " << s11n_SQLITE_TBL_T<<" where key='current_node_id'),"
			    << "'" << pkey << "',"
			    << "'" << pval << "'"
			    << ");\n";
		}
	}

	/**
	   Internal.

	   Generates SQL for saving src and sends it to os.

	   Recurses into properties and children.
	*/
	template <typename NodeType>
	void save_node_impl( NodeType const & src, std::ostream & os )
	{
		typedef s11n::node_traits<NodeType> NTR;
		std::string name = NTR::name( src ); // todo: properly escape the string
		std::string iclass = NTR::class_name( src ); // todo: properly escape the string


		std::ostringstream sql;
		os << "INSERT INTO "<<s11n_SQLITE_TBL_N<<" VALUES("
		   <<"NULL,"
		   << "ifnull((SELECT value FROM " << s11n_SQLITE_TBL_T << " WHERE key='current_node_id'),0),"
		   <<"'"<<iclass<<"',"
		   <<"'"<<name<<"'"
		   << ");\n"
			;

		// os << "SELECT 'Saving node: ' || last_insert_rowid();\n";

		typedef typename NTR::child_list_type::const_iterator CIT;
		CIT cit = NTR::children(src).begin();
		CIT cet = NTR::children(src).end();

		typedef typename NTR::property_map_type::const_iterator PIT;
		PIT pit = NTR::properties(src).begin();
		PIT pet = NTR::properties(src).end();

		bool verbose = (pit != pet) || (cet != cit);
		std::string mykey;
		if( verbose )
		{ /** we need to do extra bookwork to tell our child
		      objects our ROWID... */
			std::ostringstream lkey;
			lkey << std::hex<<&src;
			mykey = lkey.str();
			os << "REPLACE INTO "<<s11n_SQLITE_TBL_T
			   <<" VALUES('"<<mykey<<"',last_insert_rowid());\n"
				;

			os << "REPLACE INTO "<<s11n_SQLITE_TBL_T
			   <<" VALUES('current_node_id',"
			   <<"(SELECT value FROM "<<s11n_SQLITE_TBL_T
			   <<" WHERE key='"<<mykey<<"'));\n"
				;
		}

		if( pet != pit )
		{
			save_node_properties<NodeType>( src, os );
		}

		for( ; cet != cit; ++cit )
		{
			save_node_impl<NodeType>(*(*cit), os );
		}

		if( verbose )
		{
			os << "REPLACE INTO "<<s11n_SQLITE_TBL_T
			   <<" VALUES('current_node_id',"
			   <<"(SELECT parent_id FROM "<<s11n_SQLITE_TBL_N<<" WHERE rowid="
			   <<"(SELECT value FROM "<<s11n_SQLITE_TBL_T
			   <<" WHERE key='"<<mykey<<"')"
			   <<")"
			   << ");\n"
				;
		}
		return;
	}

	/**
	   Internal implementation of save_node(). Not intended for client use.

	   It recurses into src src's child objects, saving them as well.

	   src is expected to be the root object for the db. parent_id
	   should be 0 when called from other code, or a db-determined
	   value when recursed from within this function.

	   Returns the sqlite ROWID identifier associated with src or
	   throws a std::exception on error.
	*/
	template <typename NodeType>
	rowid_t save_node_impl( NodeType const & src, ::sqlite3 * db, rowid_t parent_id )
	{
		typedef s11n::node_traits<NodeType> NTR;
		const std::string name = NTR::name( src );
		const std::string iclass = NTR::class_name( src );
		//escape_string( name );
		//escape_string( iclass );

		std::ostringstream sql;
		const bool use_transaction( 0 && (0 == parent_id ) );
		if( use_transaction )
		{
			sql << "BEGIN TRANSACTION saving_root_node;\n";
		}
		sql << "INSERT INTO "<<s11n_SQLITE_TBL_N<<" values(NULL,"
		    <<parent_id<<","
		    <<"'"<<iclass<<"',"
		    <<"'"<<name<<"'"
		    << ");\n";

		int rc = // documented but nonexistant: sqlite3_exec_printf( db, sql.str().c_str(), NULL, 0, &dberr, name.c_str(), iclass.c_str() );
			sqlite3_exec( db, sql.str().c_str(), NULL, 0, 0 );
		if( SQLITE_OK != rc )
		{
			throw s11n::s11n_exception( std::string("save_node_impl(): SQL error while adding snode: ")
						    +sqlite3_errmsg(db)
						    +"\nOffending SQL code follows:\n"
						    +sql.str(),
						    __FILE__,__LINE__ );

		}
		size_t id = sqlite3_last_insert_rowid(db);
		save_node_properties<NodeType>( src, db, id );

		typedef typename NTR::child_list_type::const_iterator CIT;
		CIT it = NTR::children(src).begin();
		CIT et = NTR::children(src).end();
		for( ; et != it; ++it )
		{
			save_node_impl<NodeType>(*(*it), db, id );
		}

		if( use_transaction )
		{
			if( 0 != sqlite3_exec( db, "COMMIT TRANSACTION saving_root_node;", 0, 0, 0 ) )
			{
				throw s11n::s11n_exception( "SQL COMMIT TRANSACTION Failed.", __FILE__,__LINE__ );
			}
		}
		return id;
	}

	/**
	   Initializes db using init_db(db) and saves src into db.
	   Returns true on success, and typically throws
	   a std::exception on error. You can query
	   for the saved object by using: WHERE parent_id=0.

	   This is only to be used for saving top-level nodes.

	   This function does not check to see that the db
	   is clean for use as a target. If nodes already
	   exist in the db, this function will end up producing
	   another root node in the same db, which violates
	   the conventions of this API.
	*/
	template <typename NodeType>
	bool save_node( NodeType const & src, ::sqlite3 * db )
	{
		
		if( ! init_db(db) ) return 0;
 		sqlite3_exec( db, "BEGIN TRANSACTION saving_root_node;", 0, 0, 0 );
		rowid_t id = save_node_impl<NodeType>( src, db, 0 );
		int rc = sqlite3_exec( db, "COMMIT TRANSACTION saving_root_node;", 0, 0, 0 );
		if( 0 != rc )
		{
			throw s11n::s11n_exception( std::string("SQL Error saving root node: ")
						    +sqlite3_errmsg( db ),
						    __FILE__, __LINE__ );
		}
 		return 0 != id;
	}

	/**
	   Saves src as SQL code to dest. The SQL code can be used to
	   recreate src in a sqlite3 database. It can be *extremely*
	   verbose, and it is not uncommon for these files to be 5+
	   times the size of other s11n formats. It is, however,
	   faster than most. Also, recreating the db with the SQL is
	   much faster than saving src directly to database (not quite
	   certain why).

	   This function always returns true unless (!dest.good()).
	*/
	template <typename NodeType>
	bool save_node( NodeType const & src, std::ostream & dest )
	{
		if( ! dest.good() ) return false;
		get_header_sql( dest );
		get_drop_sql( dest );
		get_create_sql( dest );
		/**
		   WOW: adding the TRANSACTION block here cuts the
		   import time by a factor of 5 or more.
		*/
 		dest << "BEGIN TRANSACTION saving_root_node;\n";
		save_node_impl<NodeType>( src, dest );
		dest << "COMMIT TRANSACTION saving_root_node;\n";
		get_cleanup_sql( dest );
		dest << std::endl;
		return dest.good();
	}

	/**
	   Saves src to the given file as SQL. See save_node( NodeType, ostream ).
	*/
	template <typename NodeType>
	bool save_node_to_sql( NodeType const & src, std::string const & dest )
	{
		std::ofstream osql( dest.c_str() );
		if( ! osql.good() )
		{
			throw ::s11n::s11n_exception( "save_node_to_sql( node, "+dest+" ) could not open file for writing.", __FILE__, __LINE__ );
		}
		save_node<NodeType>( src, osql );
		return osql.good();
	}

	/**
	   Saves src to database file dest. If dest already exists, it
	   must be an sqlite3 database or this operation will
	   fail. See save_node(NodeType,::sqlite3*).
	*/
	template <typename NodeType>
	bool save_node_to_db( NodeType const & src, std::string const & dest )
	{
 		DBScope dbw( dest.c_str() );
 		return save_node<NodeType>( src, dbw.db );
	}

	/**
	   This variant works a bit differently than the others:

	   dest is assumed to be a filename. TWO files are generated:

	   - dest is an sqlite3 database

	   - dest.sql is the SQL code for src.

	   It returns true on success and false (or passes on an
	   exception) on error. On error, it is possible that one
	   or the other of the files was created. Their creation
	   order is not guaranteed by this function. When in doubt,
	   look for the .sql file, because its contents will make
	   tracking down the problem easier.

	   See save_node_to_sql() and save_node_to_db() for more
	   details.
	*/
	template <typename NodeType>
	bool save_node_to_both( NodeType const & src, std::string const & dest )
	{
		return save_node_to_sql(src,dest+".sql")
			&& save_node_to_db(src,dest);
	}


	/**
	   Callback for query:

	   select key,value from properties [where...]

	   cbarg is expected to be a pointer to the NodeType object associated with
	   the properties.
	*/
	template <typename NodeType>
	int properties_kv_callback(void *cbarg, int argc, char **argv, char **azColName)
	{
		typedef s11n::node_traits<NodeType> NTR;
		if( 0 == cbarg ) return 0;
		if( 2 != argc )
		{
			throw s11n::s11n_exception( "properties_kv_callback() requires that (argc==2)" );
		}
		NodeType * node = (NodeType*)cbarg; // which new-style cast is correct here?
		std::string pkey = argv[0];
		std::string pval = argv[1] ? argv[1] : "";
		NTR::set( *node, pkey, pval );
		return 0;
	}


	/**
	   Callback for query:

	   select rowid,class,name from snodes [where...]

	   cbarg is expected to be a pointer to the NodeType object
	   associated with the data, and its state is modified based
	   on the result set data.
	*/
	template <typename NodeType>
	int snodes_callback(void *cbarg, int argc, char **argv, char **azColName)
	{
		if( 0 == cbarg ) return 0;
		if( 3 != argc )
		{
			throw s11n::s11n_exception( "snodes_callback() requires that (argc==3)" );
		}
		typedef s11n::node_traits<NodeType> NTR;
		NodeType * node = (NodeType*)cbarg; // which new-style cast is correct here?
		// CERR << "snodes_callback( " << NTR::name(*node) << " )\n";
		int at = 0;
		NTR::set( *node, ".rowid", argv[at++] );
		NTR::class_name( *node, argv[at++] );
		NTR::name( *node, argv[at++] );
		return 0;
	}

	/**
	   Callback for query:

	   select rowid,class,name from snodes [where...]

	   cbarg is expected to be the parent NodeType object.
	   All child db nodes of cbarg are added to cbarg's children()
	   list. Their properties are not loaded, nor are they recursively
	   loaded, because we don't have the db handle from here.
	*/
	template <typename NodeType>
	int snodes_load_children_callback(void *cbarg, int argc, char **argv, char **azColName)
	{
		if( 0 == cbarg ) return 0;
		if( 3 != argc )
		{
			throw s11n::s11n_exception( "snodes_callback() requires that (argc==3)" );
		}
		typedef s11n::node_traits<NodeType> NTR;
		NodeType * node = (NodeType*)cbarg; // which new-style cast is correct here?
		int at = 0;
		std::auto_ptr<NodeType> ch( NTR::create() );
		NTR::set( *ch, ".rowid", argv[at++] );
		NTR::class_name( *ch, argv[at++] );
		NTR::name( *ch, argv[at++] );
		// 		CERR << "snodes_load_children_callback( " << NTR::name(*node) << " ): child ==> "
		// 		     << argv[0] << ", " << argv[1] << ", " << argv[2]
		// 		     <<"\n";
		NTR::children( *node ).push_back( ch.release() );
		return 0;
	}


	/**
	   Loads all nodes where parent_id=dest_id and adds them as
	   children of dest, recursively loading sub-nodes. Their
	   properties are loaded as well.

	   dest is expected to be a top-level snode and dest_id is
	   expected to be dest's associated sqlite ROWID value.
	*/
	template <typename NodeType>
	bool load_node_children( ::sqlite3 * db, NodeType & dest, rowid_t dest_id )
	{
		typedef s11n::node_traits<NodeType> NTR;
		// CERR << "load_node_children( " << NTR::name(dest) << " )\n";
		std::ostringstream sql;
		sql << "select rowid,class,name from "<<s11n_SQLITE_TBL_N<<" where parent_id="<<dest_id;
		SQLDEBUG << sql.str() << '\n';
		char * dberr = 0;
		int rc = sqlite3_exec( db, sql.str().c_str(), snodes_load_children_callback<NodeType>, &dest, &dberr );
		if( SQLITE_OK != rc )
		{
			throw s11n::s11n_exception( std::string("DB error: ")+dberr+"\nSQL="+sql.str() );
		}
		
		typedef typename NTR::child_list_type::iterator CLIT;
		CLIT it = NTR::children(dest).begin();
		CLIT et = NTR::children(dest).end();
		rowid_t rid = 0;
		NodeType * ch = 0;
		for( ; et != it; ++it )
		{
			ch = *it;
			rid = NTR::template get<rowid_t>( *ch, ".rowid", (rowid_t)0 );
			NTR::unset( *ch, ".rowid" );
			sql.str("");
			sql << "select key,value from "<<s11n_SQLITE_TBL_P<<" where node_id="<<rid;
			SQLDEBUG << sql.str() << '\n';
			rc = sqlite3_exec( db, sql.str().c_str(), properties_kv_callback<NodeType>, ch, &dberr );
			if( SQLITE_OK != rc )
			{
				throw s11n::s11n_exception( std::string("DB error: ")+dberr+"\nSQL="+sql.str() );
			}
			load_node_children<NodeType>( db, *ch, rid );
		}

		return true;
	}

	/**
	   Reads the root s11n object in db and returns it as a
	   NodeType tree. On error it propagates an exception or
	   returns 0. The caller owns the returned pointer.

	   ACHTUNG: Only useful for loading root nodes.
	*/
	template <typename NodeType>
	NodeType * load_node( ::sqlite3 * db )
	{
		typedef s11n::node_traits<NodeType> NTR;
		std::auto_ptr<NodeType> node( new NodeType );
		std::ostringstream sql;

		sql << "ANALYZE;"; // try to speed up loads of large data sets
		sqlite3_exec( db, sql.str().c_str(), 0, 0, 0 );
		sql.str("");
		sql << "SELECT rowid,class,name FROM "<<s11n_SQLITE_TBL_N<<" WHERE parent_id=0";
		SQLDEBUG << sql.str() << '\n';
		int rc = sqlite3_exec( db, sql.str().c_str(), snodes_callback<NodeType>, node.get(), 0 );
		if( SQLITE_OK != rc ) return 0;

		rowid_t rid = NTR::template get<rowid_t>( *node, ".rowid", (rowid_t)0 );
		NTR::unset( *node, ".rowid" );
		sql.str("");
		sql << "SELECT key,value FROM "<<s11n_SQLITE_TBL_P<<" WHERE node_id="<<rid;
		SQLDEBUG << sql.str() << '\n';
		rc = sqlite3_exec( db, sql.str().c_str(), properties_kv_callback<NodeType>, node.get(), 0 );
		if( SQLITE_OK != rc ) return 0;

		load_node_children<NodeType>( db, *node, rid );

		return node.release();
	}




} } } // namespace s11n::io::sqlite3


#undef SQLDEBUG
#endif // s11n_SQLITE3_UTIL_HPP_INCLUDED
